#Turtle.py - Written by James Reinebold (www.reinebold.com)



#imports
import math
import pygame

#constants
DEGREES2RADIANS = 0.01745

class Turtle:
    """
        Turtle drawing class for Pygame.
        Renders the l-system string by parsing it as a series of commands.
        All drawing is done from the perspective of the turtle.
        See http://en.wikipedia.org/wiki/L-system for more information.
    """

    def __init__(self, rules, myX, myY, myAngle, mySpeed):
        """
            Create a new turtle.

            Rules - must contain a control string
            Rules - must contain a dictionary that explains how to draw:
            The config dictionary is a lookup table where the keys are the characters
            and the values are turtle drawing commands (as explained in the readme).
           
            myX - initial x position on screen for turtle
            myY - initial y position on screen for turtle
            myAngle - initial degree facing for turtle (90 is up)
            mySpeed - how many pixels this turtle can draw per second
        """
        self.string = rules.string
        self.x = myX
        self.y = myY
        self.angle = myAngle
        self.stack = []
        self.config = rules.config
        self.speed = mySpeed #how many pixels to draw per second
        self.pixels2Draw = 0.0
        self.pixelsDrawnOnLine = 0.0
        self.current = 0
        self.doneLines = []
        self.color = [0, 100, 0]

    def update(self, dt):
        """
            Figure out how many pixels we can draw this round.
            dt = in milliseconds
        """
        timeInSeconds = dt / 1000.0
        #round up because we always want to draw at least one pixel
        self.pixels2Draw += round(self.speed * timeInSeconds + 0.5)

    def draw(self, display):
        """
            Draw to the screen.
            Will gradually draw the lsystem based on how many pixels of it
            we are allowed to draw per frame.
            When the whole l system has been formed, keep drawing it.
        """
        for line in self.doneLines:
            pygame.draw.aaline(display, self.color, line[0], line[1])
        if self.current >= len(self.string):
            pass
        else:
            while self.pixels2Draw > 0.0001 and self.current < len(self.string):
                char = self.string[self.current]
                action = self.config[char]
                change = self.handleAction(action, display)
                if change:
                    self.current += 1

    def drawLinePart(self, display, pos1, pos2):
        """
            Draw so many pixels of the line.
            If numPixels > size of line, draw entire line.
            Return true if the entire line was drawn
            Return false if only part of the line was drawn
        """
        lineDist = math.sqrt(math.pow(pos1[0] - pos2[0], 2) + math.pow(pos1[1] - pos2[1], 2))
        self.pixelsDrawnOnLine = min(self.pixels2Draw + self.pixelsDrawnOnLine, lineDist)
        if lineDist <= self.pixelsDrawnOnLine:
            #draw the whole line
            pygame.draw.aaline(display, self.color, pos1, pos2)
            self.pixels2Draw -= lineDist
            return True
        else:
            #draw only how many pixels we are permitted to along the direction vector of the line
            direction = [pos2[0] - pos1[0], pos2[1] - pos1[0]]
            mag = math.sqrt(math.pow(direction[0], 2) + math.pow(direction[1], 2))
            direction[0] /= mag
            direction[1] /= mag
            direction[0] *= self.pixelsDrawnOnLine / lineDist
            direction[1] *= self.pixelsDrawnOnLine / lineDist
            pygame.draw.aaline(display, self.color, pos1, [pos1[0] + direction[0], pos1[1] + direction[1]])
            return False

    def handleAction(self, action, display):
        """
            Handle one character of the string.
            Return True if we fully "handled" the character this round.
        """
        if action.startswith("PUSH"):
            #push to the stack
            self.stack.append((self.x, self.y, self.angle))
            return True
        elif action.startswith("POP"):
            #pop from the stack
            data = self.stack.pop()
            self.x = data[0]
            self.y = data[1]
            self.angle = data[2]
            return True
        elif action.startswith("NULL"):
            #do nothing
            return True
        elif action.startswith("TURN"):
            #turn the facing of the turtle
            parts = action.split()
            otherAngle = float(parts[1])
            self.angle += otherAngle
            return True
        elif action.startswith("DRAW"):
            #draw straight along the direction the turtle is facing
            parts = action.split()
            distance = float(parts[1])
            compX = math.cos(self.angle * DEGREES2RADIANS) * distance
            compY = math.sin(self.angle * DEGREES2RADIANS) * -1 * distance
            newX = compX + self.x
            newY = compY + self.y
            done = self.drawLinePart(display, [self.x, self.y], [newX, newY])
            self.pixels2Draw -= self.pixelsDrawnOnLine
            if done:
                #we drew the whole line
                self.doneLines.append([[self.x, self.y], [newX, newY]])
                self.x = newX
                self.y = newY
                return True
            #we will have to finish drawing the line later
            return False
